/* Authors: Lutong Wang and Bangqi Xu */
/*
 * Copyright (c) 2019, The Regents of the University of California
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <algorithm>
#include <iterator>
#include <map>
#include <memory>
#include <utility>

#include "db/tech/frLookupTbl.h"
#include "frBaseTypes.h"
#include "frViaDef.h"
#include "frViaRuleGenerate.h"
#include "odb/db.h"
#include "utl/Logger.h"

namespace fr {
class frLayer;
namespace io {
class Parser;
}

using ForbiddenRanges = std::vector<std::pair<frCoord, frCoord>>;

enum class frLef58CornerSpacingExceptionEnum
{
  NONE,
  EXCEPTSAMENET,
  EXCEPTSAMEMETAL
};

struct drEolSpacingConstraint
{
  drEolSpacingConstraint(frCoord width = 0,
                         frCoord space = 0,
                         frCoord within = 0)
      : eolWidth(width), eolSpace(space), eolWithin(within)
  {
  }
  frCoord eolWidth;
  frCoord eolSpace;
  frCoord eolWithin;
};

// base type for design rule
class frConstraint
{
 public:
  virtual ~frConstraint() {}
  virtual frConstraintTypeEnum typeId() const = 0;
  virtual void report(utl::Logger* logger) const = 0;
  void setLayer(frLayer* layer) { layer_ = layer; }
  void setId(int in) { id_ = in; }
  int getId() const { return id_; }
  std::string getViolName() const
  {
    switch (typeId()) {
      case frConstraintTypeEnum::frcShortConstraint:
        return "Short";
      case frConstraintTypeEnum::frcMinWidthConstraint:
        return "Min Width";

      case frConstraintTypeEnum::frcSpacingConstraint:
        return "Metal Spacing";

      case frConstraintTypeEnum::frcSpacingEndOfLineConstraint:
        return "EOL Spacing";

      case frConstraintTypeEnum::frcSpacingTablePrlConstraint:
        return "Metal Spacing";

      case frConstraintTypeEnum::frcCutSpacingConstraint:
        return "Cut Spacing";

      case frConstraintTypeEnum::frcMinStepConstraint:
        return "Min Step";

      case frConstraintTypeEnum::frcNonSufficientMetalConstraint:
        return "NS Metal";

      case frConstraintTypeEnum::frcSpacingSamenetConstraint:
        return "Metal Spacing";

      case frConstraintTypeEnum::frcOffGridConstraint:
        return "Off Grid";

      case frConstraintTypeEnum::frcMinEnclosedAreaConstraint:
        return "Min Hole";

      case frConstraintTypeEnum::frcAreaConstraint:
        return "Min Area";

      case frConstraintTypeEnum::frcLef58CornerSpacingConstraint:
        return "Corner Spacing";

      case frConstraintTypeEnum::frcLef58CutSpacingConstraint:
        return "Cut Spacing";

      case frConstraintTypeEnum::frcLef58RectOnlyConstraint:
        return "Rect Only";

      case frConstraintTypeEnum::frcLef58RightWayOnGridOnlyConstraint:
        return "RightWayOnGridOnly";

      case frConstraintTypeEnum::frcLef58MinStepConstraint:
        return "Min Step";

      case frConstraintTypeEnum::frcSpacingTableInfluenceConstraint:
        return "MetSpacingInf";

      case frConstraintTypeEnum::frcSpacingEndOfLineParallelEdgeConstraint:
        return "SpacingEOLParallelEdge";

      case frConstraintTypeEnum::frcSpacingTableConstraint:
        return "SpacingTable";

      case frConstraintTypeEnum::frcSpacingTableTwConstraint:
        return "SpacingTableTw";

      case frConstraintTypeEnum::frcLef58SpacingTableConstraint:
        return "Lef58SpacingTable";

      case frConstraintTypeEnum::frcLef58CutSpacingTableConstraint:
        return "Lef58CutSpacingTable";

      case frConstraintTypeEnum::frcLef58CutSpacingTablePrlConstraint:
        return "Lef58CutSpacingTablePrl";

      case frConstraintTypeEnum::frcLef58CutSpacingTableLayerConstraint:
        return "Lef58CutSpacingTableLayer";

      case frConstraintTypeEnum::frcLef58CutSpacingParallelWithinConstraint:
        return "Lef58CutSpacingParallelWithin";

      case frConstraintTypeEnum::frcLef58CutSpacingAdjacentCutsConstraint:
        return "Lef58CutSpacingAdjacentCuts";

      case frConstraintTypeEnum::frcLef58CutSpacingLayerConstraint:
        return "Lef58CutSpacingLayer";

      case frConstraintTypeEnum::frcMinimumcutConstraint:
      case frConstraintTypeEnum::frcLef58MinimumCutConstraint:
        return "Minimum Cut";

      case frConstraintTypeEnum::frcLef58CornerSpacingConcaveCornerConstraint:
        return "Lef58CornerSpacingConcaveCorner";

      case frConstraintTypeEnum::frcLef58CornerSpacingConvexCornerConstraint:
        return "Lef58CornerSpacingConvexCorner";

      case frConstraintTypeEnum::frcLef58CornerSpacingSpacingConstraint:
        return "Lef58CornerSpacingSpacing";

      case frConstraintTypeEnum::frcLef58CornerSpacingSpacing1DConstraint:
        return "Lef58CornerSpacingSpacing1D";

      case frConstraintTypeEnum::frcLef58CornerSpacingSpacing2DConstraint:
        return "Lef58CornerSpacingSpacing2D";

      case frConstraintTypeEnum::frcLef58SpacingEndOfLineConstraint:
        return "Lef58SpacingEndOfLine";

      case frConstraintTypeEnum::frcLef58SpacingEndOfLineWithinConstraint:
        return "Lef58SpacingEndOfLineWithin";

      case frConstraintTypeEnum::
          frcLef58SpacingEndOfLineWithinEndToEndConstraint:
        return "Lef58SpacingEndOfLineWithinEndToEnd";

      case frConstraintTypeEnum::
          frcLef58SpacingEndOfLineWithinEncloseCutConstraint:
        return "Lef58SpacingEndOfLineWithinEncloseCut";

      case frConstraintTypeEnum::
          frcLef58SpacingEndOfLineWithinParallelEdgeConstraint:
        return "Lef58SpacingEndOfLineWithinParallelEdge";

      case frConstraintTypeEnum::
          frcLef58SpacingEndOfLineWithinMaxMinLengthConstraint:
        return "Lef58SpacingEndOfLineWithinMaxMinLength";

      case frConstraintTypeEnum::frcLef58SpacingWrongDirConstraint:
        return "Lef58SpacingWrongDir";

      case frConstraintTypeEnum::frcLef58CutClassConstraint:
        return "Lef58CutClass";

      case frConstraintTypeEnum::frcRecheckConstraint:
        return "Recheck";

      case frConstraintTypeEnum::frcLef58EolExtensionConstraint:
        return "Lef58EolExtension";

      case frConstraintTypeEnum::frcLef58EolKeepOutConstraint:
        return "Lef58EolKeepOut";

      case frConstraintTypeEnum::frcMetalWidthViaConstraint:
        return "MetalWidthViaMap";

      case frConstraintTypeEnum::frcLef58AreaConstraint:
        return "Lef58Area";
      case frConstraintTypeEnum::frcLef58KeepOutZoneConstraint:
        return "KeepOutZone";
      case frConstraintTypeEnum::frcSpacingRangeConstraint:
        return "SpacingRange";
    }
    return "";
  }

 protected:
  int id_;
  frLayer* layer_;
  frConstraint() : id_(-1), layer_(nullptr) {}
};

class frLef58CutClassConstraint : public frConstraint
{
 public:
  friend class io::Parser;
  // constructors;
  frLef58CutClassConstraint() {}
  // getters
  frCollection<std::shared_ptr<frLef58CutClass>> getCutClasses() const
  {
    frCollection<std::shared_ptr<frLef58CutClass>> sol;
    std::transform(cutClasses.begin(),
                   cutClasses.end(),
                   std::back_inserter(sol),
                   [](auto& kv) { return kv.second; });
    return sol;
  }
  // setters
  void addToCutClass(const std::shared_ptr<frLef58CutClass>& in)
  {
    cutClasses[in->getName()] = in;
  }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58CutClassConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("Cut class");
  }

 protected:
  std::map<frString, std::shared_ptr<frLef58CutClass>> cutClasses;
};

// recheck constraint for negative rules
class frRecheckConstraint : public frConstraint
{
 public:
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcRecheckConstraint;
  }
  void report(utl::Logger* logger) const override { logger->report("Recheck"); }
};

// short

class frShortConstraint : public frConstraint
{
 public:
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcShortConstraint;
  }
  void report(utl::Logger* logger) const override { logger->report("Short"); }
};

// NSMetal
class frNonSufficientMetalConstraint : public frConstraint
{
 public:
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcNonSufficientMetalConstraint;
  }
  void report(utl::Logger* logger) const override { logger->report("NSMetal"); }
};

// offGrid
class frOffGridConstraint : public frConstraint
{
 public:
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcOffGridConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("Off grid");
  }
};

// minHole
class frMinEnclosedAreaConstraint : public frConstraint
{
 public:
  // constructor
  frMinEnclosedAreaConstraint(frCoord areaIn) : area(areaIn), width(-1) {}
  // getter
  frCoord getArea() const { return area; }
  bool hasWidth() const { return (width != -1); }
  frCoord getWidth() const { return width; }
  // setter
  void setWidth(frCoord widthIn) { width = widthIn; }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcMinEnclosedAreaConstraint;
  }

  void report(utl::Logger* logger) const override
  {
    logger->report("Min enclosed area {} width {}", area, width);
  }

 protected:
  frCoord area, width;
};

// LEF58_MINSTEP (currently only implement GF14 related API)
class frLef58MinStepConstraint : public frConstraint
{
 public:
  // constructor
  frLef58MinStepConstraint()
      : minStepLength(-1),
        insideCorner(false),
        outsideCorner(false),
        step(false),
        maxLength(-1),
        maxEdges(-1),
        minAdjLength(-1),
        convexCorner(false),
        exceptWithin(-1),
        concaveCorner(false),
        threeConcaveCorners(false),
        width(-1),
        minAdjLength2(-1),
        minBetweenLength(-1),
        exceptSameCorners(false),
        eolWidth(-1),
        concaveCorners(false)
  {
  }
  // getter
  frCoord getMinStepLength() const { return minStepLength; }
  bool hasMaxEdges() const { return (maxEdges != -1); }
  int getMaxEdges() const { return maxEdges; }
  bool hasMinAdjacentLength() const { return (minAdjLength != -1); }
  frCoord getMinAdjacentLength() const { return minAdjLength; }
  bool hasEolWidth() const { return (eolWidth != -1); }
  frCoord getEolWidth() const { return eolWidth; }

  // setter
  void setMinStepLength(frCoord in) { minStepLength = in; }
  void setMaxEdges(int in) { maxEdges = in; }
  void setMinAdjacentLength(frCoord in) { minAdjLength = in; }
  void setEolWidth(frCoord in) { eolWidth = in; }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58MinStepConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report(
        "MINSTEP minStepLength {} insideCorner {} outsideCorner {} step {} "
        "maxLength {} maxEdges {} minAdjLength {} convexCorner {} exceptWithin "
        "{} concaveCorner {} threeConcaveCorners {} width {} minAdjLength2 {} "
        "minBetweenLength {} exceptSameCorners {} eolWidth {} concaveCorners "
        "{} ",
        minStepLength,
        insideCorner,
        outsideCorner,
        step,
        maxLength,
        maxEdges,
        minAdjLength,
        convexCorner,
        exceptWithin,
        concaveCorner,
        threeConcaveCorners,
        width,
        minAdjLength2,
        minBetweenLength,
        exceptSameCorners,
        eolWidth,
        concaveCorners);
  }

 protected:
  frCoord minStepLength;
  bool insideCorner;
  bool outsideCorner;
  bool step;
  frCoord maxLength;
  int maxEdges;
  frCoord minAdjLength;
  bool convexCorner;
  frCoord exceptWithin;
  bool concaveCorner;
  bool threeConcaveCorners;
  frCoord width;
  frCoord minAdjLength2;
  frCoord minBetweenLength;
  bool exceptSameCorners;
  frCoord eolWidth;
  bool concaveCorners;
};

// minStep
class frMinStepConstraint : public frConstraint
{
 public:
  // constructor
  frMinStepConstraint()
      : minStepLength(-1),
        minstepType(frMinstepTypeEnum::UNKNOWN),
        maxLength(-1),
        insideCorner(false),
        outsideCorner(true),
        step(false),
        maxEdges(-1)
  {
  }
  // getter
  frCoord getMinStepLength() const { return minStepLength; }
  bool hasMaxLength() const { return (maxLength != -1); }
  frCoord getMaxLength() const { return maxLength; }
  bool hasMinstepType() const
  {
    return minstepType != frMinstepTypeEnum::UNKNOWN;
  }
  frMinstepTypeEnum getMinstepType() const { return minstepType; }
  bool hasInsideCorner() const { return insideCorner; }
  bool hasOutsideCorner() const { return outsideCorner; }
  bool hasStep() const { return step; }
  bool hasMaxEdges() const { return (maxEdges != -1); }
  int getMaxEdges() const { return maxEdges; }
  // setter
  void setMinstepType(frMinstepTypeEnum in) { minstepType = in; }
  void setInsideCorner(bool in) { insideCorner = in; }
  void setOutsideCorner(bool in) { outsideCorner = in; }
  void setStep(bool in) { step = in; }
  void setMinStepLength(frCoord in) { minStepLength = in; }
  void setMaxLength(frCoord in) { maxLength = in; }
  void setMaxEdges(int in) { maxEdges = in; }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcMinStepConstraint;
  }

  void report(utl::Logger* logger) const override
  {
    logger->report(
        "Min step length min {} type {} max {} "
        "insideCorner {} outsideCorner {} step {} maxEdges {}",
        minStepLength,
        int(minstepType),
        maxLength,
        insideCorner,
        outsideCorner,
        step,
        maxEdges);
  }

 protected:
  frCoord minStepLength;
  frMinstepTypeEnum minstepType;
  frCoord maxLength;
  bool insideCorner;
  bool outsideCorner;
  bool step;
  int maxEdges;
};

// minimumcut
class frMinimumcutConstraint : public frConstraint
{
 public:
  frMinimumcutConstraint()
      : numCuts(-1),
        width(-1),
        cutDistance(-1),
        connection(frMinimumcutConnectionEnum::UNKNOWN),
        length(-1),
        distance(-1)
  {
  }
  // getters
  int getNumCuts() const { return numCuts; }
  frCoord getWidth() const { return width; }
  bool hasWithin() const { return !(cutDistance == -1); }
  frCoord getCutDistance() const { return cutDistance; }
  bool hasConnection() const
  {
    return !(connection == frMinimumcutConnectionEnum::UNKNOWN);
  }
  frMinimumcutConnectionEnum getConnection() const { return connection; }
  bool hasLength() const { return !(length == -1); }
  frCoord getLength() const { return length; }
  frCoord getDistance() const { return distance; }
  // setters
  void setNumCuts(int in) { numCuts = in; }
  void setWidth(frCoord in) { width = in; }
  void setWithin(frCoord in) { cutDistance = in; }
  void setConnection(frMinimumcutConnectionEnum in) { connection = in; }
  void setLength(frCoord in1, frCoord in2)
  {
    length = in1;
    distance = in2;
  }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcMinimumcutConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report(
        "Min cut numCuts {} width {} cutDistance {} "
        "connection {} length {} distance {}",
        numCuts,
        width,
        cutDistance,
        connection,
        length,
        distance);
  }

 protected:
  int numCuts;
  frCoord width;
  frCoord cutDistance;
  frMinimumcutConnectionEnum connection;
  frCoord length;
  frCoord distance;
};

// LEF58_MINIMUMCUT
class frLef58MinimumcutConstraint : public frConstraint
{
 public:
  frLef58MinimumcutConstraint(odb::dbTechLayerMinCutRule* rule) : db_rule_(rule)
  {
  }
  // getter
  odb::dbTechLayerMinCutRule* getODBRule() const { return db_rule_; }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58MinimumCutConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("LEF58_MINIMUMCUT");
  }

 private:
  odb::dbTechLayerMinCutRule* db_rule_;
};

// minArea

class frAreaConstraint : public frConstraint
{
 public:
  // constructor
  frAreaConstraint(frCoord minAreaIn) { minArea = minAreaIn; }
  // getter
  frCoord getMinArea() { return minArea; }
  // setter
  void setMinArea(frCoord minAreaIn) { minArea = minAreaIn; }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcAreaConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("Area {}", minArea);
  }

 protected:
  frCoord minArea;
};

// minWidth
class frMinWidthConstraint : public frConstraint
{
 public:
  // constructor
  frMinWidthConstraint(frCoord minWidthIn) { minWidth = minWidthIn; }
  // getter
  frCoord getMinWidth() { return minWidth; }
  // setter
  void set(frCoord minWidthIn) { minWidth = minWidthIn; }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcMinWidthConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("Width {}", minWidth);
  }

 protected:
  frCoord minWidth;
};

class frLef58SpacingEndOfLineWithinEncloseCutConstraint : public frConstraint
{
 public:
  // constructors
  frLef58SpacingEndOfLineWithinEncloseCutConstraint(frCoord encloseDistIn,
                                                    frCoord cutToMetalSpaceIn)
      : below(false),
        above(false),
        encloseDist(encloseDistIn),
        cutToMetalSpace(cutToMetalSpaceIn),
        allCuts(false)
  {
  }
  // setters
  void setBelow(bool value) { below = value; }
  void setAbove(bool value) { above = value; }
  void setAllCuts(bool value) { allCuts = value; }
  void setEncloseDist(frCoord value) { encloseDist = value; }
  void setCutToMetalSpace(frCoord value) { cutToMetalSpace = value; }
  // getters
  bool isAboveOnly() const { return above; }
  bool isBelowOnly() const { return below; }
  bool isAboveAndBelow() const { return !(above ^ below); }
  bool isAllCuts() const { return allCuts; }
  frCoord getEncloseDist() const { return encloseDist; }
  frCoord getCutToMetalSpace() const { return cutToMetalSpace; }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::
        frcLef58SpacingEndOfLineWithinEncloseCutConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report(
        "\t\tSPACING_WITHIN_ENCLOSECUT below {} above {} encloseDist "
        "{} cutToMetalSpace {} allCuts {}",
        below,
        above,
        encloseDist,
        cutToMetalSpace,
        allCuts);
  }

 private:
  bool below;
  bool above;
  frCoord encloseDist;
  frCoord cutToMetalSpace;
  bool allCuts;
};

class frLef58SpacingEndOfLineWithinEndToEndConstraint : public frConstraint
{
 public:
  // constructors
  frLef58SpacingEndOfLineWithinEndToEndConstraint()
      : endToEndSpace(0),
        cutSpace(false),
        oneCutSpace(0),
        twoCutSpace(0),
        hExtension(false),
        extension(0),
        wrongDirExtension(false),
        hOtherEndWidth(false),
        otherEndWidth(0)
  {
  }
  // getters
  frCoord getEndToEndSpace() const { return endToEndSpace; }
  frCoord getOneCutSpace() const { return oneCutSpace; }
  frCoord getTwoCutSpace() const { return twoCutSpace; }
  bool hasExtension() const { return hExtension; }
  frCoord getExtension() const { return extension; }
  frCoord getWrongDirExtension() const { return wrongDirExtension; }
  bool hasOtherEndWidth() const { return hOtherEndWidth; }
  frCoord getOtherEndWidth() const { return otherEndWidth; }

  // setters
  void setEndToEndSpace(frCoord in) { endToEndSpace = in; }
  void setCutSpace(frCoord one, frCoord two)
  {
    oneCutSpace = one;
    twoCutSpace = two;
  }
  void setExtension(frCoord extensionIn)
  {
    hExtension = true;
    extension = extensionIn;
    wrongDirExtension = extensionIn;
  }
  void setExtension(frCoord extensionIn, frCoord wrongDirExtensionIn)
  {
    hExtension = true;
    extension = extensionIn;
    wrongDirExtension = wrongDirExtensionIn;
  }
  void setOtherEndWidth(frCoord in)
  {
    hOtherEndWidth = true;
    otherEndWidth = in;
  }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::
        frcLef58SpacingEndOfLineWithinEndToEndConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report(
        "\t\tSPACING_WITHIN_ENDTOEND endToEndSpace {} cutSpace {} oneCutSpace "
        "{} twoCutSpace {} hExtension {} extension {} wrongDirExtension {} "
        "hOtherEndWidth {} otherEndWidth {} ",
        endToEndSpace,
        cutSpace,
        oneCutSpace,
        twoCutSpace,
        hExtension,
        extension,
        wrongDirExtension,
        hOtherEndWidth,
        otherEndWidth);
  }

 protected:
  frCoord endToEndSpace;
  bool cutSpace;
  frCoord oneCutSpace;
  frCoord twoCutSpace;
  bool hExtension;
  frCoord extension;
  frCoord wrongDirExtension;
  bool hOtherEndWidth;
  frCoord otherEndWidth;
};

class frLef58SpacingEndOfLineWithinParallelEdgeConstraint : public frConstraint
{
 public:
  // constructors
  frLef58SpacingEndOfLineWithinParallelEdgeConstraint()
      : subtractEolWidth(false),
        parSpace(0),
        parWithin(0),
        hPrl(false),
        prl(0),
        hMinLength(false),
        minLength(0),
        twoEdges(false),
        sameMetal(false),
        nonEolCornerOnly(false),
        parallelSameMask(false)
  {
  }
  // getters
  bool hasSubtractEolWidth() const { return subtractEolWidth; }
  frCoord getParSpace() const { return parSpace; }
  frCoord getParWithin() const { return parWithin; }
  bool hasPrl() const { return hPrl; }
  frCoord getPrl() const { return prl; }
  bool hasMinLength() const { return hMinLength; }
  frCoord getMinLength() const { return minLength; }
  bool hasTwoEdges() const { return twoEdges; }
  bool hasSameMetal() const { return sameMetal; }
  bool hasNonEolCornerOnly() const { return nonEolCornerOnly; }
  bool hasParallelSameMask() const { return parallelSameMask; }
  // setters
  void setSubtractEolWidth(bool in) { subtractEolWidth = in; }
  void setPar(frCoord parSpaceIn, frCoord parWithinIn)
  {
    parSpace = parSpaceIn;
    parWithin = parWithinIn;
  }
  void setPrl(frCoord in)
  {
    hPrl = true;
    prl = in;
  }
  void setMinLength(frCoord in)
  {
    hMinLength = true;
    minLength = in;
  }
  void setTwoEdges(bool in) { twoEdges = in; }
  void setSameMetal(bool in) { sameMetal = in; }
  void setNonEolCornerOnly(bool in) { nonEolCornerOnly = in; }
  void setParallelSameMask(bool in) { parallelSameMask = in; }

  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::
        frcLef58SpacingEndOfLineWithinParallelEdgeConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report(
        "\t\tSPACING_WITHIN_PARALLELEDGE subtractEolWidth {} parSpace {} "
        "parWithin {} hPrl {} prl {} hMinLength {} minLength {} twoEdges {} "
        "sameMetal {} nonEolCornerOnly {} parallelSameMask {} ",
        subtractEolWidth,
        parSpace,
        parWithin,
        hPrl,
        prl,
        hMinLength,
        minLength,
        twoEdges,
        sameMetal,
        nonEolCornerOnly,
        parallelSameMask);
  }

 protected:
  bool subtractEolWidth;
  frCoord parSpace;
  frCoord parWithin;
  bool hPrl;
  frCoord prl;
  bool hMinLength;
  frCoord minLength;
  bool twoEdges;
  bool sameMetal;
  bool nonEolCornerOnly;
  bool parallelSameMask;
};

class frLef58SpacingEndOfLineWithinMaxMinLengthConstraint : public frConstraint
{
 public:
  // constructors
  frLef58SpacingEndOfLineWithinMaxMinLengthConstraint()
      : maxLength(false), length(0), twoSides(false)
  {
  }

  // getters
  frCoord getLength() const { return length; }
  bool isMaxLength() const { return maxLength; }
  bool isTwoSides() const { return twoSides; }

  // setters
  void setLength(bool maxLengthIn, frCoord lengthIn, bool twoSidesIn = false)
  {
    maxLength = maxLengthIn;
    length = lengthIn;
    twoSides = twoSidesIn;
  }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::
        frcLef58SpacingEndOfLineWithinMaxMinLengthConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report(
        "\t\tSPACING_WITHIN_MAXMIN maxLength {} length {} twoSides {} ",
        maxLength,
        length,
        twoSides);
  }

 protected:
  bool maxLength;
  frCoord length;
  bool twoSides;
};

class frLef58SpacingEndOfLineWithinConstraint : public frConstraint
{
 public:
  // constructors
  frLef58SpacingEndOfLineWithinConstraint()
      : hOppositeWidth(false),
        oppositeWidth(0),
        eolWithin(0),
        wrongDirWithin(false),
        endPrlSpacing(0),
        endPrl(0),
        sameMask(false),
        endToEndConstraint(nullptr),
        parallelEdgeConstraint(nullptr)
  {
  }

  // getters
  bool hasOppositeWidth() const { return hOppositeWidth; }
  frCoord getOppositeWidth() const { return oppositeWidth; }
  frCoord getEolWithin() const { return sameMask ? 0 : eolWithin; }
  frCoord getWrongDirWithin() const { return wrongDirWithin; }
  frCoord getEndPrlSpacing() const { return endPrlSpacing; }
  frCoord getEndPrl() const { return endPrl; }
  bool hasSameMask() const { return sameMask; }
  bool hasExceptExactWidth() const
  {
    return false;  // skip for now
  }
  bool hasFillConcaveCorner() const
  {
    return false;  // skip for now
  }
  bool hasWithCut() const
  {
    return false;  // skip for now
  }
  bool hasEndPrlSpacing() const { return endPrlSpacing; }
  bool hasEndToEndConstraint() const
  {
    return (endToEndConstraint) ? true : false;
  }
  std::shared_ptr<frLef58SpacingEndOfLineWithinEndToEndConstraint>
  getEndToEndConstraint() const
  {
    return endToEndConstraint;
  }
  bool hasMinMaxLength() const
  {
    return false;  // skip for now
  }
  bool hasEqualRectWidth() const
  {
    return false;  // skip for now
  }
  bool hasParallelEdgeConstraint() const
  {
    return (parallelEdgeConstraint) ? true : false;
  }
  std::shared_ptr<frLef58SpacingEndOfLineWithinParallelEdgeConstraint>
  getParallelEdgeConstraint() const
  {
    return parallelEdgeConstraint;
  }
  bool hasMaxMinLengthConstraint() const
  {
    return (maxMinLengthConstraint) ? true : false;
  }
  std::shared_ptr<frLef58SpacingEndOfLineWithinMaxMinLengthConstraint>
  getMaxMinLengthConstraint() const
  {
    return maxMinLengthConstraint;
  }
  bool hasEncloseCutConstraint() const
  {
    return (encloseCutConstraint) ? true : false;
  }
  std::shared_ptr<frLef58SpacingEndOfLineWithinEncloseCutConstraint>
  getEncloseCutConstraint() const
  {
    return encloseCutConstraint;
  }
  // setters
  void setOppositeWidth(frCoord in)
  {
    hOppositeWidth = true;
    oppositeWidth = in;
  }
  void setEolWithin(frCoord in)
  {
    eolWithin = in;
    wrongDirWithin = in;
  }
  void setEndPrl(frCoord endPrlSpacingIn, frCoord endPrlIn)
  {
    endPrlSpacing = endPrlSpacingIn;
    endPrl = endPrlIn;
  }
  void setWrongDirWithin(frCoord in) { wrongDirWithin = in; }
  void setSameMask(bool in) { sameMask = in; }
  void setEndToEndConstraint(
      const std::shared_ptr<frLef58SpacingEndOfLineWithinEndToEndConstraint>&
          in)
  {
    endToEndConstraint = in;
  }
  void setParallelEdgeConstraint(
      const std::shared_ptr<
          frLef58SpacingEndOfLineWithinParallelEdgeConstraint>& in)
  {
    parallelEdgeConstraint = in;
  }
  void setMaxMinLengthConstraint(
      const std::shared_ptr<
          frLef58SpacingEndOfLineWithinMaxMinLengthConstraint>& in)
  {
    maxMinLengthConstraint = in;
  }
  void setEncloseCutConstraint(
      const std::shared_ptr<frLef58SpacingEndOfLineWithinEncloseCutConstraint>&
          in)
  {
    encloseCutConstraint = in;
  }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58SpacingEndOfLineWithinConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report(
        "\tSPACING_WITHIN hOppositeWidth {} oppositeWidth {} eolWithin {} "
        "wrongDirWithin {} sameMask {} endPrlSpacing {} endPrl {}",
        hOppositeWidth,
        oppositeWidth,
        eolWithin,
        wrongDirWithin,
        sameMask,
        endPrlSpacing,
        endPrl);
    if (endToEndConstraint != nullptr)
      endToEndConstraint->report(logger);
    if (parallelEdgeConstraint != nullptr)
      parallelEdgeConstraint->report(logger);
  }

 protected:
  bool hOppositeWidth;
  frCoord oppositeWidth;
  frCoord eolWithin;
  frCoord wrongDirWithin;
  frCoord endPrlSpacing;
  frCoord endPrl;
  bool sameMask;
  std::shared_ptr<frLef58SpacingEndOfLineWithinEndToEndConstraint>
      endToEndConstraint;
  std::shared_ptr<frLef58SpacingEndOfLineWithinParallelEdgeConstraint>
      parallelEdgeConstraint;
  std::shared_ptr<frLef58SpacingEndOfLineWithinMaxMinLengthConstraint>
      maxMinLengthConstraint;
  std::shared_ptr<frLef58SpacingEndOfLineWithinEncloseCutConstraint>
      encloseCutConstraint;
};

class frLef58SpacingEndOfLineConstraint : public frConstraint
{
 public:
  // constructors
  frLef58SpacingEndOfLineConstraint()
      : eolSpace(0),
        eolWidth(0),
        exactWidth(false),
        wrongDirSpacing(false),
        wrongDirSpace(0),
        withinConstraint(nullptr)
  {
  }
  // getters
  frCoord getEolSpace() const { return eolSpace; }
  frCoord getEolWidth() const { return eolWidth; }
  bool hasExactWidth() const { return exactWidth; }
  bool hasWrongDirSpacing() const { return wrongDirSpacing; }
  frCoord getWrongDirSpace() const { return wrongDirSpace; }
  bool hasWithinConstraint() const { return (withinConstraint) ? true : false; }
  std::shared_ptr<frLef58SpacingEndOfLineWithinConstraint> getWithinConstraint()
      const
  {
    return withinConstraint;
  }
  bool hasToConcaveCornerConstraint() const { return false; }
  bool hasToNotchLengthConstraint() const { return false; }
  // setters
  void setEol(frCoord eolSpaceIn, frCoord eolWidthIn, bool exactWidthIn = false)
  {
    eolSpace = eolSpaceIn;
    eolWidth = eolWidthIn;
    exactWidth = exactWidthIn;
  }
  void setWrongDirSpace(frCoord in)
  {
    wrongDirSpacing = true;
    wrongDirSpace = in;
  }
  void setWithinConstraint(
      const std::shared_ptr<frLef58SpacingEndOfLineWithinConstraint>& in)
  {
    withinConstraint = in;
  }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58SpacingEndOfLineConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report(
        "SPACING eolSpace {} eolWidth {} exactWidth {} wrongDirSpacing {} "
        "wrongDirSpace {} ",
        eolSpace,
        eolWidth,
        exactWidth,
        wrongDirSpacing,
        wrongDirSpace);
    if (withinConstraint != nullptr)
      withinConstraint->report(logger);
  }

 protected:
  frCoord eolSpace;
  frCoord eolWidth;
  bool exactWidth;
  bool wrongDirSpacing;
  frCoord wrongDirSpace;
  std::shared_ptr<frLef58SpacingEndOfLineWithinConstraint> withinConstraint;
};

class frLef58SpacingWrongDirConstraint : public frConstraint
{
 public:
  frLef58SpacingWrongDirConstraint(odb::dbTechLayerWrongDirSpacingRule* rule)
      : db_rule_(rule)
  {
  }
  // getter
  odb::dbTechLayerWrongDirSpacingRule* getODBRule() const { return db_rule_; }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58SpacingWrongDirConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("LEF58_SPACING WRONGDIRECTION");
  }

 private:
  odb::dbTechLayerWrongDirSpacingRule* db_rule_;
};

class frLef58EolKeepOutConstraint : public frConstraint
{
 public:
  // constructors
  frLef58EolKeepOutConstraint()
      : backwardExt(0),
        sideExt(0),
        forwardExt(0),
        eolWidth(0),
        cornerOnly(false),
        exceptWithin(false),
        withinLow(0),
        withinHigh(0)
  {
  }
  // getters
  frCoord getBackwardExt() const { return backwardExt; }
  frCoord getForwardExt() const { return forwardExt; }
  frCoord getSideExt() const { return sideExt; }
  frCoord getEolWidth() const { return eolWidth; }
  bool isCornerOnly() const { return cornerOnly; }
  bool isExceptWithin() const { return exceptWithin; }
  frCoord getWithinLow() const { return withinLow; }
  frCoord getWithinHigh() const { return withinHigh; }
  // setters
  void setBackwardExt(frCoord value) { backwardExt = value; }
  void setForwardExt(frCoord value) { forwardExt = value; }
  void setSideExt(frCoord value) { sideExt = value; }
  void setEolWidth(frCoord value) { eolWidth = value; }
  void setCornerOnly(bool value) { cornerOnly = value; }
  void setExceptWithin(bool value) { exceptWithin = value; }
  void setWithinLow(frCoord value) { withinLow = value; }
  void setWithinHigh(frCoord value) { withinHigh = value; }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58EolKeepOutConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report(
        "EOLKEEPOUT backwardExt {} sideExt {} forwardExt {} eolWidth {} "
        "cornerOnly {} exceptWithin {} withinLow {} withinHigh {}",
        backwardExt,
        sideExt,
        forwardExt,
        eolWidth,
        cornerOnly,
        exceptWithin,
        withinLow,
        withinHigh);
  }

 private:
  frCoord backwardExt;
  frCoord sideExt;
  frCoord forwardExt;
  frCoord eolWidth;
  bool cornerOnly;
  bool exceptWithin;
  frCoord withinLow, withinHigh;
};

class frLef58CornerSpacingSpacingConstraint;

// SPACING Constraints
class frSpacingConstraint : public frConstraint
{
 public:
  frSpacingConstraint() : minSpacing(0) {}
  frSpacingConstraint(frCoord minSpacingIn) : minSpacing(minSpacingIn) {}

  // getter
  frCoord getMinSpacing() const { return minSpacing; }
  // setter
  void setMinSpacing(frCoord minSpacingIn) { minSpacing = minSpacingIn; }
  // check
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcSpacingConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("Spacing {}", minSpacing);
  }

 protected:
  frCoord minSpacing;
};

class frSpacingSamenetConstraint : public frSpacingConstraint
{
 public:
  frSpacingSamenetConstraint() : frSpacingConstraint(), pgonly(false) {}
  frSpacingSamenetConstraint(frCoord minSpacingIn, bool pgonlyIn)
      : frSpacingConstraint(minSpacingIn), pgonly(pgonlyIn)
  {
  }
  // getter
  bool hasPGonly() const { return pgonly; }
  // setter
  void setPGonly(bool in) { pgonly = in; }
  // check
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcSpacingSamenetConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("Spacing same net pgonly {}", pgonly);
  }

 protected:
  bool pgonly;
};

class frSpacingTableInfluenceConstraint : public frConstraint
{
 public:
  frSpacingTableInfluenceConstraint(
      const fr1DLookupTbl<frCoord, std::pair<frCoord, frCoord>>& in)
      : tbl(in)
  {
  }
  // getter
  const fr1DLookupTbl<frCoord, std::pair<frCoord, frCoord>>& getLookupTbl()
      const
  {
    return tbl;
  }
  std::pair<frCoord, frCoord> find(frCoord width) const
  {
    return tbl.find(width);
  }
  frCoord getMinWidth() const { return tbl.getMinRow(); }
  // setter
  void setLookupTbl(
      const fr1DLookupTbl<frCoord, std::pair<frCoord, frCoord>>& in)
  {
    tbl = in;
  }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcSpacingTableInfluenceConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("Spacing table influence");
  }

 private:
  fr1DLookupTbl<frCoord, std::pair<frCoord, frCoord>> tbl;
};
// range spacing
class frSpacingRangeConstraint : public frSpacingConstraint
{
 public:
  // constructor
  frSpacingRangeConstraint() {}
  // getters
  frCoord getMinWidth() const { return minWidth; }
  frCoord getMaxWidth() const { return minWidth; }
  // setters
  void setMinWidth(frCoord in) { minWidth = in; }
  void setMaxWidth(frCoord in) { maxWidth = in; }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcSpacingRangeConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("Spacing RANGE minWidth {} maxWidth {}", minWidth, maxWidth);
  }
  bool inRange(frCoord width) const
  {
    return width >= minWidth && width <= maxWidth;
  }

 private:
  frCoord minWidth{0};
  frCoord maxWidth{0};
};

//
// EOL spacing
class frSpacingEndOfLineConstraint : public frSpacingConstraint
{
 public:
  // constructor
  frSpacingEndOfLineConstraint()
      : frSpacingConstraint(),
        eolWidth(-1),
        eolWithin(-1),
        parSpace(-1),
        parWithin(-1),
        isTwoEdges(false)
  {
  }
  // getter
  frCoord getEolWidth() const { return eolWidth; }
  frCoord getEolWithin() const { return eolWithin; }
  frCoord getParSpace() const { return parSpace; }
  frCoord getParWithin() const { return parWithin; }
  bool hasParallelEdge() const { return ((parSpace == -1) ? false : true); }
  bool hasTwoEdges() const { return isTwoEdges; }
  // setter
  void setEolWithin(frCoord eolWithinIn) { eolWithin = eolWithinIn; }
  void setEolWidth(frCoord eolWidthIn) { eolWidth = eolWidthIn; }
  void setParSpace(frCoord parSpaceIn) { parSpace = parSpaceIn; }
  void setParWithin(frCoord parWithinIn) { parWithin = parWithinIn; }
  void setTwoEdges(bool isTwoEdgesIn) { isTwoEdges = isTwoEdgesIn; }
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcSpacingEndOfLineConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report(
        "Spacing EOL eolWidth {} eolWithin {} "
        "parSpace {} parWithin {} isTwoEdges {}",
        eolWidth,
        eolWithin,
        parSpace,
        parWithin,
        isTwoEdges);
  }

 protected:
  frCoord eolWidth, eolWithin;
  frCoord parSpace, parWithin;
  bool isTwoEdges;
};

class frLef58EolExtensionConstraint : public frSpacingConstraint
{
 public:
  // constructors
  frLef58EolExtensionConstraint(const fr1DLookupTbl<frCoord, frCoord>& tbl)
      : frSpacingConstraint(), parallelOnly(false), extensionTbl(tbl)
  {
  }
  // setters

  void setParallelOnly(bool value) { parallelOnly = value; }

  // getters

  bool isParallelOnly() const { return parallelOnly; }

  fr1DLookupTbl<frCoord, frCoord> getExtensionTable() const
  {
    return extensionTbl;
  }

  // others

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58EolExtensionConstraint;
  }

  void report(utl::Logger* logger) const override
  {
    logger->report("EOLEXTENSIONSPACING spacing {} parallelonly {} ",
                   minSpacing,
                   parallelOnly);
  }

 private:
  bool parallelOnly;
  fr1DLookupTbl<frCoord, frCoord> extensionTbl;
};

// LEF58 cut spacing table
class frLef58CutSpacingTableConstraint : public frConstraint
{
 public:
  // constructor
  frLef58CutSpacingTableConstraint(
      odb::dbTechLayerCutSpacingTableDefRule* dbRule)
      : db_rule_(dbRule),
        default_spacing_({0, 0}),
        default_center2center_(false),
        default_centerAndEdge_(false)
  {
  }
  // setter
  void setDefaultSpacing(const std::pair<frCoord, frCoord>& value)
  {
    default_spacing_ = value;
  }
  void setDefaultCenterToCenter(bool value) { default_center2center_ = value; }
  void setDefaultCenterAndEdge(bool value) { default_centerAndEdge_ = value; }
  // getter
  odb::dbTechLayerCutSpacingTableDefRule* getODBRule() const
  {
    return db_rule_;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report(
        "CUTSPACINGTABLE lyr:{} lyr2:{}",
        db_rule_->getTechLayer()->getName(),
        db_rule_->isLayerValid() ? db_rule_->getSecondLayer()->getName() : "-");
  }
  std::pair<frCoord, frCoord> getDefaultSpacing() const
  {
    return default_spacing_;
  }
  bool getDefaultCenterToCenter() const { return default_center2center_; }
  bool getDefaultCenterAndEdge() const { return default_centerAndEdge_; }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58CutSpacingTableConstraint;
  }

 private:
  odb::dbTechLayerCutSpacingTableDefRule* db_rule_;
  std::pair<frCoord, frCoord> default_spacing_;
  bool default_center2center_, default_centerAndEdge_;
};

// new SPACINGTABLE Constraints
class frSpacingTablePrlConstraint : public frConstraint
{
 public:
  // constructor
  frSpacingTablePrlConstraint(
      const fr2DLookupTbl<frCoord, frCoord, frCoord>& in)
      : tbl(in)
  {
  }
  // getter
  const fr2DLookupTbl<frCoord, frCoord, frCoord>& getLookupTbl() const
  {
    return tbl;
  }
  frCoord find(frCoord width, frCoord prl) const
  {
    return tbl.find(width, prl);
  }
  frCoord findMin() const { return tbl.findMin(); }
  frCoord findMax() const { return tbl.findMax(); }
  // setter
  void setLookupTbl(const fr2DLookupTbl<frCoord, frCoord, frCoord>& in)
  {
    tbl = in;
  }
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcSpacingTablePrlConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("Spacing table PRL");
  }

 protected:
  fr2DLookupTbl<frCoord, frCoord, frCoord> tbl;
};

struct frSpacingTableTwRowType
{
  frSpacingTableTwRowType(frCoord in1, frCoord in2) : width(in1), prl(in2) {}
  frCoord width;
  frCoord prl;
};

// new SPACINGTABLE Constraints
class frSpacingTableTwConstraint : public frConstraint
{
 public:
  // constructor
  frSpacingTableTwConstraint(
      const frCollection<frSpacingTableTwRowType>& rowsIn,
      const frCollection<frCollection<frCoord>>& spacingIn)
      : rows(rowsIn), spacingTbl(spacingIn)
  {
  }
  // getter
  frCoord find(frCoord width1, frCoord width2, frCoord prl) const;
  frCoord findMin() const { return spacingTbl.front().front(); }
  frCoord findMax() const { return spacingTbl.back().back(); }
  // setter
  void setSpacingTable(const frCollection<frSpacingTableTwRowType>& rowsIn,
                       const frCollection<frCollection<frCoord>>& spacingIn)
  {
    rows = rowsIn;
    spacingTbl = spacingIn;
  }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcSpacingTableTwConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("Spacing table tw");
  }

 private:
  frCollection<frSpacingTableTwRowType> rows;
  frCollection<frCollection<frCoord>> spacingTbl;

  frUInt4 getIdx(frCoord width, frCoord prl) const
  {
    int sz = rows.size();
    for (int i = 0; i < sz; i++) {
      if (width <= rows[i].width)
        return std::max(0, i - 1);
      if (rows[i].prl != -1 && prl <= rows[i].prl)
        return std::max(0, i - 1);
    }
    return sz - 1;
  }
};

// original SPACINGTABLE Constraints
class frSpacingTableConstraint : public frConstraint
{
 public:
  // constructor
  frSpacingTableConstraint(
      std::shared_ptr<fr2DLookupTbl<frCoord, frCoord, frCoord>>
          parallelRunLengthConstraintIn)
  {
    parallelRunLengthConstraint = parallelRunLengthConstraintIn;
  }
  // getter
  const std::shared_ptr<fr2DLookupTbl<frCoord, frCoord, frCoord>>
  getParallelRunLengthConstraint() const
  {
    return parallelRunLengthConstraint;
  }
  // setter
  void setParallelRunLengthConstraint(
      std::shared_ptr<fr2DLookupTbl<frCoord, frCoord, frCoord>>&
          parallelRunLengthConstraintIn)
  {
    parallelRunLengthConstraint = parallelRunLengthConstraintIn;
  }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcSpacingTableConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("Spacing table");
  }

 protected:
  std::shared_ptr<fr2DLookupTbl<frCoord, frCoord, frCoord>>
      parallelRunLengthConstraint;
};

class frLef58SpacingTableConstraint : public frSpacingTableConstraint
{
 public:
  // constructor
  frLef58SpacingTableConstraint(
      const std::shared_ptr<fr2DLookupTbl<frCoord, frCoord, frCoord>>&
          parallelRunLengthConstraintIn,
      const std::map<int, std::pair<frCoord, frCoord>>&
          exceptWithinConstraintIn)
      : frSpacingTableConstraint(parallelRunLengthConstraintIn),
        exceptWithinConstraint(exceptWithinConstraintIn),
        wrongDirection(false),
        sameMask(false),
        exceptEol(false),
        eolWidth(0)
  {
  }
  // getter
  bool hasExceptWithin(frCoord val) const
  {
    auto rowIdx = getParallelRunLengthConstraint()->getRowIdx(val);
    return (exceptWithinConstraint.find(rowIdx)
            != exceptWithinConstraint.end());
  }
  std::pair<frCoord, frCoord> getExceptWithin(frCoord val) const
  {
    auto rowIdx = getParallelRunLengthConstraint()->getRowIdx(val);
    return exceptWithinConstraint.at(rowIdx);
  }
  bool isWrongDirection() const { return wrongDirection; }
  bool isSameMask() const { return sameMask; }
  bool hasExceptEol() const { return exceptEol; }
  frUInt4 getEolWidth() const { return eolWidth; }
  // setters
  void setExceptWithinConstraint(
      std::map<int, std::pair<frCoord, frCoord>>& exceptWithinConstraintIn)
  {
    exceptWithinConstraint = exceptWithinConstraintIn;
  }
  void setWrongDirection(bool in) { wrongDirection = in; }
  void setSameMask(bool in) { sameMask = in; }
  void setEolWidth(frUInt4 in)
  {
    exceptEol = true;
    eolWidth = in;
  }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58SpacingTableConstraint;
  }

  void report(utl::Logger* logger) const override
  {
    logger->report(
        "SPACINGTABLE wrongDirection {} sameMask {} exceptEol {} eolWidth {} ",
        wrongDirection,
        sameMask,
        exceptEol,
        eolWidth);
    logger->report("\texceptWithinConstraint");
    for (auto& [key, val] : exceptWithinConstraint)
      logger->report("\t{} ({} {})", key, val.first, val.second);
  }

 protected:
  std::map<frCoord, std::pair<frCoord, frCoord>> exceptWithinConstraint;
  bool wrongDirection;
  bool sameMask;
  bool exceptEol;
  frUInt4 eolWidth;
};

// ADJACENTCUTS
class frCutSpacingConstraint : public frConstraint
{
 public:
  // constructor
  frCutSpacingConstraint() {}
  frCutSpacingConstraint(frCoord cutSpacingIn,
                         bool centerToCenterIn,
                         bool sameNetIn,
                         frString secondLayerNameIn,
                         bool stackIn,
                         int adjacentCutsIn,
                         frCoord cutWithinIn,
                         bool isExceptSamePGNetIn,
                         bool isParallelOverlapIn,
                         frCoord cutAreaIn,
                         int twoCutsIn = -1)
  {
    cutSpacing = cutSpacingIn;
    centerToCenter = centerToCenterIn;
    sameNet = sameNetIn;
    secondLayerName = secondLayerNameIn;
    stack = stackIn;
    adjacentCuts = adjacentCutsIn;
    cutWithin = cutWithinIn;
    exceptSamePGNet = isExceptSamePGNetIn;
    parallelOverlap = isParallelOverlapIn;
    cutArea = cutAreaIn;
    twoCuts = twoCutsIn;
  }
  // getter
  bool hasCenterToCenter() const { return centerToCenter; }
  bool getCenterToCenter() const { return centerToCenter; }
  bool getSameNet() const { return sameNet; }
  bool hasSameNet() const { return sameNet; }
  frCutSpacingConstraint* getSameNetConstraint() { return sameNetConstraint; }
  bool getStack() const { return stack; }
  bool hasStack() const { return stack; }
  bool isLayer() const { return !(secondLayerName.empty()); }
  const frString& getSecondLayerName() const { return secondLayerName; }
  bool hasSecondLayer() const { return (secondLayerNum != -1); }
  frLayerNum getSecondLayerNum() const { return secondLayerNum; }
  bool isAdjacentCuts() const { return (adjacentCuts != -1); }
  int getAdjacentCuts() const { return adjacentCuts; }
  frCoord getCutWithin() const { return cutWithin; }
  bool hasExceptSamePGNet() const { return exceptSamePGNet; }
  bool getExceptSamePGNet() const { return exceptSamePGNet; }
  bool isParallelOverlap() const { return parallelOverlap; }
  bool getParallelOverlap() const { return parallelOverlap; }
  bool isArea() const { return !(cutArea == -1); }
  bool getCutArea() const { return cutArea; }
  frCoord getCutSpacing() const { return cutSpacing; }
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcCutSpacingConstraint;
  }
  // LEF58 related
  bool isTwoCuts() const { return (twoCuts == -1); }
  int getTwoCuts() const { return twoCuts; }

  void report(utl::Logger* logger) const override
  {
    logger->report("Cut Spacing");
  }

  // setter

  void setSecondLayerNum(int secondLayerNumIn)
  {
    secondLayerNum = secondLayerNumIn;
  }

  void setSameNetConstraint(frCutSpacingConstraint* in)
  {
    sameNetConstraint = in;
  }

 protected:
  frCoord cutSpacing = -1;
  bool centerToCenter = false;
  bool sameNet = false;
  frCutSpacingConstraint* sameNetConstraint = nullptr;
  bool stack = false;
  bool exceptSamePGNet = false;
  bool parallelOverlap = false;
  frString secondLayerName;
  frLayerNum secondLayerNum = -1;
  int adjacentCuts = -1;
  frCoord cutWithin = -1;
  frCoord cutArea = -1;
  // LEF58 related
  int twoCuts = -1;
};

// LEF58_SPACING for cut layer (new)
class frLef58CutSpacingConstraint : public frConstraint
{
 public:
  // constructor
  frLef58CutSpacingConstraint()
      : cutSpacing(-1),
        sameMask(false),
        maxXY(false),
        centerToCenter(false),
        sameNet(false),
        sameMetal(false),
        sameVia(false),
        secondLayerName(""),
        secondLayerNum(-1),
        stack(false),
        orthogonalSpacing(-1),
        cutClassName(""),
        cutClassIdx(-1),
        shortEdgeOnly(false),
        prl(-1),
        concaveCorner(false),
        width(-1),
        enclosure(-1),
        edgeLength(-1),
        parLength(-1),
        parWithin(-1),
        edgeEnclosure(-1),
        adjEnclosure(-1),
        extension(-1),
        eolWidth(-1),
        minLength(-1),
        maskOverlap(false),
        wrongDirection(false),
        adjacentCuts(-1),
        exactAlignedCut(-1),
        twoCuts(-1),
        twoCutsSpacing(-1),
        sameCut(false),
        cutWithin1(-1),
        cutWithin2(-1),
        exceptSamePGNet(false),
        exceptAllWithin(-1),
        above(false),
        below(false),
        toAll(false),
        noPrl(false),
        sideParallelOverlap(false),
        parallelOverlap(false),
        exceptSameNet(false),
        exceptSameMetal(false),
        exceptSameMetalOverlap(false),
        exceptSameVia(false),
        within(-1),
        longEdgeOnly(false),
        exceptTwoEdges(false),
        numCut(-1),
        cutArea(-1)
  {
  }
  // getters
  // is == what rules have; has == what derived from values
  frCoord getCutSpacing() const { return cutSpacing; }
  bool isSameMask() const { return sameMask; }
  bool isMaxXY() const { return maxXY; }
  bool isCenterToCenter() const { return centerToCenter; }
  bool isSameNet() const { return sameNet; }
  bool isSameMetal() const { return sameMetal; }
  bool isSameVia() const { return sameVia; }
  std::string getSecondLayerName() const { return secondLayerName; }
  bool hasSecondLayer() const
  {
    return (secondLayerNum != -1 || secondLayerName != std::string(""));
  }
  frLayerNum getSecondLayerNum() const { return secondLayerNum; }
  bool isStack() const { return stack; }
  bool hasOrthogonalSpacing() const { return (orthogonalSpacing != -1); }
  std::string getCutClassName() const { return cutClassName; }
  bool hasCutClass() const
  {
    return (cutClassIdx != -1 || cutClassName != std::string(""));
  }
  int getCutClassIdx() const { return cutClassIdx; }
  bool isShortEdgeOnly() const { return shortEdgeOnly; }
  bool hasPrl() const { return (prl != -1); }
  frCoord getPrl() const { return prl; }
  bool isConcaveCorner() const { return concaveCorner; }
  bool hasWidth() const { return (width != -1); }
  frCoord getWidth() const { return width; }
  bool hasEnclosure() const { return (enclosure != -1); }
  frCoord getEnclosure() const { return enclosure; }
  bool hasEdgeLength() const { return (edgeLength != -1); }
  frCoord getEdgeLength() const { return edgeLength; }
  bool hasParallel() const { return (parLength != -1); }
  frCoord getParlength() const { return parLength; }
  frCoord getParWithin() const { return parWithin; }
  frCoord getEdgeEnclosure() const { return edgeEnclosure; }
  frCoord getAdjEnclosure() const { return adjEnclosure; }
  bool hasExtension() const { return (extension != -1); }
  frCoord getExtension() const { return extension; }
  bool hasNonEolConvexCorner() const { return (eolWidth != -1); }
  frCoord getEolWidth() const { return eolWidth; }
  bool hasMinLength() const { return (minLength != -1); }
  frCoord getMinLength() const { return minLength; }
  bool hasAboveWidth() const { return (width != -1); }
  bool isMaskOverlap() const { return maskOverlap; }
  bool isWrongDirection() const { return wrongDirection; }
  bool hasAdjacentCuts() const { return (adjacentCuts != -1); }
  int getAdjacentCuts() const { return adjacentCuts; }
  bool hasExactAligned() const { return (exactAlignedCut != -1); }
  int getExactAligned() const { return exactAlignedCut; }
  bool hasTwoCuts() const { return (twoCuts != -1); }
  int getTwoCuts() const { return twoCuts; }
  bool hasTwoCutsSpacing() const { return (twoCutsSpacing != -1); }
  frCoord getTwoCutsSpacing() const { return twoCutsSpacing; }
  bool isSameCut() const { return sameCut; }
  // cutWithin2 is always used as upper bound
  bool hasTwoCutWithin() const { return (cutWithin1 != -1); }
  frCoord getCutWithin() const { return cutWithin2; }
  frCoord getCutWithin1() const { return cutWithin1; }
  frCoord getCutWithin2() const { return cutWithin2; }
  bool isExceptSamePGNet() const { return exceptSamePGNet; }
  bool hasExceptAllWithin() const { return (exceptAllWithin != -1); }
  frCoord getExceptAllWithin() const { return exceptAllWithin; }
  bool isAbove() const { return above; }
  bool isBelow() const { return below; }
  bool isToAll() const { return toAll; }
  bool isNoPrl() const { return noPrl; }
  bool isSideParallelOverlap() const { return sideParallelOverlap; }
  bool isParallelOverlap() const { return parallelOverlap; }
  bool isExceptSameNet() const { return exceptSameNet; }
  bool isExceptSameMetal() const { return exceptSameMetal; }
  bool isExceptSameMetalOverlap() const { return exceptSameMetalOverlap; }
  bool isExceptSameVia() const { return exceptSameVia; }
  bool hasParallelWithin() const { return (within != -1); }
  frCoord getWithin() const { return within; }
  bool isLongEdgeOnly() const { return longEdgeOnly; }
  bool hasSameMetalSharedEdge() const { return (parWithin != -1); }
  bool isExceptTwoEdges() const { return exceptTwoEdges; }
  bool hasExceptSameVia() const { return (numCut != -1); }
  bool hasArea() const { return (cutArea != -1); }
  frCoord getCutArea() const { return cutArea; }
  // setters
  void setCutSpacing(frCoord in) { cutSpacing = in; }
  void setSameMask(bool in) { sameMask = in; }
  void setMaxXY(bool in) { maxXY = in; }
  void setCenterToCenter(bool in) { centerToCenter = in; }
  void setSameNet(bool in) { sameNet = in; }
  void setSameMetal(bool in) { sameMetal = in; }
  void setSameVia(bool in) { sameVia = in; }
  void setSecondLayerName(const std::string& in) { secondLayerName = in; }
  void setSecondLayerNum(frLayerNum in) { secondLayerNum = in; }
  void setStack(bool in) { stack = in; }
  void setOrthogonalSpacing(frCoord in) { orthogonalSpacing = in; }
  void setCutClassName(const std::string& in) { cutClassName = in; }
  void setCutClassIdx(int in) { cutClassIdx = in; }
  void setShortEdgeOnly(bool in) { shortEdgeOnly = in; }
  void setPrl(frCoord in) { prl = in; }
  void setConcaveCorner(bool in) { concaveCorner = in; }
  void setWidth(frCoord in) { width = in; }
  void setEnclosure(frCoord in) { enclosure = in; }
  void setEdgeLength(frCoord in) { edgeLength = in; }
  void setParLength(frCoord in) { parLength = in; }
  void setParWithin(frCoord in) { parWithin = in; }
  void setEdgeEnclosure(frCoord in) { edgeEnclosure = in; }
  void setAdjEnclosure(frCoord in) { adjEnclosure = in; }
  void setExtension(frCoord in) { extension = in; }
  void setEolWidth(frCoord in) { eolWidth = in; }
  void setMinLength(frCoord in) { minLength = in; }
  void setMaskOverlap(bool in) { maskOverlap = in; }
  void setWrongDirection(bool in) { wrongDirection = in; }
  void setAdjacentCuts(int in) { adjacentCuts = in; }
  void setExactAlignedCut(int in) { exactAlignedCut = in; }
  void setTwoCuts(int in) { twoCuts = in; }
  void setTwoCutsSpacing(frCoord in) { twoCutsSpacing = in; }
  void setSameCut(bool in) { sameCut = in; }
  void setCutWithin(frCoord in) { cutWithin2 = in; }
  void setCutWithin1(frCoord in) { cutWithin1 = in; }
  void setCutWithin2(frCoord in) { cutWithin2 = in; }
  void setExceptSamePGNet(bool in) { exceptSamePGNet = in; }
  void setExceptAllWithin(frCoord in) { exceptAllWithin = in; }
  void setAbove(bool in) { above = in; }
  void setBelow(bool in) { below = in; }
  void setToAll(bool in) { toAll = in; }
  void setNoPrl(bool in) { noPrl = in; }
  void setSideParallelOverlap(bool in) { sideParallelOverlap = in; }
  void setParallelOverlap(bool in) { parallelOverlap = in; }
  void setExceptSameNet(bool in) { exceptSameNet = in; }
  void setExceptSameMetal(bool in) { exceptSameMetal = in; }
  void setExceptSameMetalOverlap(bool in) { exceptSameMetalOverlap = in; }
  void setExceptSameVia(bool in) { exceptSameVia = in; }
  void setWithin(frCoord in) { within = in; }
  void setLongEdgeOnly(bool in) { longEdgeOnly = in; }
  void setExceptTwoEdges(bool in) { exceptTwoEdges = in; }
  void setNumCut(int in) { numCut = in; }
  void setCutArea(frCoord in) { cutArea = in; }

  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58CutSpacingConstraint;
  }

  void report(utl::Logger* logger) const override
  {
    logger->report(
        "CUTSPACING cutSpacing {} sameMask {} maxXY {} centerToCenter {} "
        "sameNet {} sameMetal {} sameVia {} secondLayerName {} secondLayerNum "
        "{} stack {} orthogonalSpacing {} cutClassName {} cutClassIdx {} "
        "shortEdgeOnly {} prl {} concaveCorner {} width {} enclosure {} "
        "edgeLength {} parLength {} parWithin {} edgeEnclosure {} adjEnclosure "
        "{} extension {} eolWidth {} minLength {} maskOverlap {} "
        "wrongDirection {} adjacentCuts {} exactAlignedCut {} twoCuts {} "
        "twoCutsSpacing {} sameCut {} cutWithin1 {} cutWithin2 {} "
        "exceptSamePGNet {} exceptAllWithin {} above {} below {} toAll {} "
        "noPrl {} sideParallelOverlap {} parallelOverlap {} exceptSameNet {} "
        "exceptSameMetal {} exceptSameMetalOverlap {} exceptSameVia {} within "
        "{} longEdgeOnly {} exceptTwoEdges {} numCut {} cutArea {} ",
        cutSpacing,
        sameMask,
        maxXY,
        centerToCenter,
        sameNet,
        sameMetal,
        sameVia,
        secondLayerName,
        secondLayerNum,
        stack,
        orthogonalSpacing,
        cutClassName,
        cutClassIdx,
        shortEdgeOnly,
        prl,
        concaveCorner,
        width,
        enclosure,
        edgeLength,
        parLength,
        parWithin,
        edgeEnclosure,
        adjEnclosure,
        extension,
        eolWidth,
        minLength,
        maskOverlap,
        wrongDirection,
        adjacentCuts,
        exactAlignedCut,
        twoCuts,
        twoCutsSpacing,
        sameCut,
        cutWithin1,
        cutWithin2,
        exceptSamePGNet,
        exceptAllWithin,
        above,
        below,
        toAll,
        noPrl,
        sideParallelOverlap,
        parallelOverlap,
        exceptSameNet,
        exceptSameMetal,
        exceptSameMetalOverlap,
        exceptSameVia,
        within,
        longEdgeOnly,
        exceptTwoEdges,
        numCut,
        cutArea);
  }

 protected:
  frCoord cutSpacing;
  bool sameMask;
  bool maxXY;
  bool centerToCenter;
  bool sameNet;
  bool sameMetal;
  bool sameVia;
  std::string secondLayerName;
  frLayerNum secondLayerNum;
  bool stack;
  frCoord orthogonalSpacing;
  std::string cutClassName;
  int cutClassIdx;
  bool shortEdgeOnly;
  frCoord prl;
  bool concaveCorner;
  frCoord width;
  frCoord enclosure;
  frCoord edgeLength;
  frCoord parLength;
  frCoord parWithin;
  frCoord edgeEnclosure;
  frCoord adjEnclosure;
  frCoord extension;
  frCoord eolWidth;
  frCoord minLength;
  bool maskOverlap;
  bool wrongDirection;
  int adjacentCuts;
  int exactAlignedCut;
  int twoCuts;
  frCoord twoCutsSpacing;
  bool sameCut;
  frCoord cutWithin1;
  frCoord cutWithin2;
  bool exceptSamePGNet;
  frCoord exceptAllWithin;
  bool above;
  bool below;
  bool toAll;
  bool noPrl;
  bool sideParallelOverlap;
  bool parallelOverlap;
  bool exceptSameNet;
  bool exceptSameMetal;
  bool exceptSameMetalOverlap;
  bool exceptSameVia;
  frCoord within;
  bool longEdgeOnly;
  bool exceptTwoEdges;
  int numCut;
  frCoord cutArea;
};

// LEF58_CORNERSPACING (new)
class frLef58CornerSpacingConstraint : public frConstraint
{
 public:
  // constructor
  frLef58CornerSpacingConstraint(
      const fr1DLookupTbl<frCoord, std::pair<frCoord, frCoord>>& tblIn)
      : cornerType(frCornerTypeEnum::UNKNOWN),
        sameMask(false),
        within(-1),
        eolWidth(-1),
        length(-1),
        edgeLength(false),
        includeLShape(false),
        minLength(-1),
        exceptNotch(false),
        notchLength(-1),
        exceptSameNet(false),
        exceptSameMetal(false),
        tbl(tblIn),
        sameXY(false),
        cornerToCorner(false)
  {
  }

  // getters
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58CornerSpacingConstraint;
  }
  frCornerTypeEnum getCornerType() const { return cornerType; }
  bool getSameMask() const { return sameMask; }
  bool hasCornerOnly() const { return (within != -1); }
  frCoord getWithin() const { return within; }
  bool hasExceptEol() const { return (eolWidth != -1); }
  frCoord getEolWidth() const { return eolWidth; }
  bool hasExceptJogLength() const { return (minLength != -1); }
  frCoord getLength() const { return length; }
  bool hasEdgeLength() const { return edgeLength; }
  bool getEdgeLength() const { return edgeLength; }
  bool hasIncludeLShape() const { return includeLShape; }
  bool getIncludeLShape() const { return includeLShape; }
  frCoord getMinLength() const { return minLength; }
  bool hasExceptNotch() const { return exceptNotch; }
  bool getExceptNotch() const { return exceptNotch; }
  frCoord getNotchLength() const { return notchLength; }
  bool hasExceptSameNet() const { return exceptSameNet; }
  bool getExceptSameNet() const { return exceptSameNet; }
  bool hasExceptSameMetal() const { return exceptSameMetal; }
  bool getExceptSameMetal() const { return exceptSameMetal; }
  frCoord find(frCoord width, bool isHorizontal = true) const
  {
    return (isHorizontal ? tbl.find(width).first : tbl.find(width).second);
  }
  frCoord findMin() const
  {
    return std::min(tbl.findMin().first, tbl.findMin().second);
  }
  frCoord findMax() const
  {
    return std::max(tbl.findMax().first, tbl.findMax().second);
  }
  frCoord findMax(bool isHorz) const
  {
    return (isHorz ? tbl.findMax().first : tbl.findMax().second);
  }
  bool hasSameXY() const { return sameXY; }
  bool getSameXY() const { return sameXY; }
  bool isCornerToCorner() const { return cornerToCorner; }

  // setters
  void setCornerType(frCornerTypeEnum in) { cornerType = in; }
  void setSameMask(bool in) { sameMask = in; }
  void setWithin(frCoord in) { within = in; }
  void setEolWidth(frCoord in) { eolWidth = in; }
  void setLength(frCoord in) { length = in; }
  void setEdgeLength(bool in) { edgeLength = in; }
  void setIncludeLShape(bool in) { includeLShape = in; }
  void setMinLength(frCoord in) { minLength = in; }
  void setExceptNotch(bool in) { exceptNotch = in; }
  void setExceptNotchLength(frCoord in) { notchLength = in; }
  void setExceptSameNet(bool in) { exceptSameNet = in; }
  void setExceptSameMetal(bool in) { exceptSameMetal = in; }
  void setLookupTbl(
      const fr1DLookupTbl<frCoord, std::pair<frCoord, frCoord>>& in)
  {
    tbl = in;
  }
  void setSameXY(bool in) { sameXY = in; }
  void setCornerToCorner(bool in) { cornerToCorner = in; }
  void report(utl::Logger* logger) const override
  {
    logger->report(
        "CORNERSPACING cornerType {} sameMask {} within {} eolWidth {} length "
        "{} edgeLength {} includeLShape {} minLength {} exceptNotch {} "
        "notchLength {} exceptSameNet {} exceptSameMetal {} sameXY {} "
        "cornerToCorner {}",
        cornerType,
        sameMask,
        within,
        eolWidth,
        length,
        edgeLength,
        includeLShape,
        minLength,
        exceptNotch,
        notchLength,
        exceptSameNet,
        exceptSameMetal,
        sameXY,
        cornerToCorner);

    std::string vals = "";
    std::string rows = "";
    for (auto row : tbl.getRows())
      rows = rows + std::to_string(row) + " ";
    for (const auto& val : tbl.getValues())
      vals = vals + "(" + std::to_string(val.first) + ","
             + std::to_string(val.second) + ") ";
    logger->report("\trowName: {}", tbl.getRowName());
    logger->report("\trows: {}", rows);
    logger->report("\tvals: {}", vals);
  }

 protected:
  frCornerTypeEnum cornerType;
  bool sameMask;
  frCoord within;
  frCoord eolWidth;
  frCoord length;
  bool edgeLength;
  bool includeLShape;
  frCoord minLength;
  bool exceptNotch;
  frCoord notchLength;
  bool exceptSameNet;
  bool exceptSameMetal;
  fr1DLookupTbl<frCoord, std::pair<frCoord, frCoord>>
      tbl;      // horz / vert spacing
  bool sameXY;  // indicate whether horz spacing == vert spacing // for write
                // LEF some day
  bool cornerToCorner;
};

class frLef58CornerSpacingSpacingConstraint : public frConstraint
{
 public:
  // constructor
  frLef58CornerSpacingSpacingConstraint(frCoord widthIn) : width(widthIn) {}
  // getter
  frCoord getWidth() { return width; }
  // setter
  void setWidth(frCoord widthIn) { width = widthIn; }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58CornerSpacingSpacingConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("58 Corner spacing spacing {}", width);
  }

 protected:
  frCoord width;
};

class frLef58CornerSpacingSpacing1DConstraint
    : public frLef58CornerSpacingSpacingConstraint
{
 public:
  // constructor
  frLef58CornerSpacingSpacing1DConstraint(frCoord widthIn, frCoord spacingIn)
      : frLef58CornerSpacingSpacingConstraint(widthIn), spacing(spacingIn)
  {
  }
  // getter
  frCoord getSpacing() { return spacing; }
  // setter
  void setSpacing(frCoord spacingIn) { spacing = spacingIn; }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58CornerSpacingSpacing1DConstraint;
  }

  void report(utl::Logger* logger) const override
  {
    logger->report("58 Corner spacing 1D {}", spacing);
  }

 protected:
  frCoord spacing = -1;
};

class frLef58CornerSpacingSpacing2DConstraint
    : public frLef58CornerSpacingSpacingConstraint
{
 public:
  // constructor
  frLef58CornerSpacingSpacing2DConstraint(frCoord widthIn,
                                          frCoord horizontalSpacingIn,
                                          frCoord verticalSpacingIn)
      : frLef58CornerSpacingSpacingConstraint(widthIn),
        horizontalSpacing(horizontalSpacingIn),
        verticalSpacing(verticalSpacingIn)
  {
  }
  // getter
  frCoord getHorizontalSpacing() { return horizontalSpacing; }
  frCoord getVerticalSpacing() { return verticalSpacing; }
  // setter
  void setHorizontalSpacing(frCoord horizontalSpacingIn)
  {
    horizontalSpacing = horizontalSpacingIn;
  }
  void setVerticalSpacing(frCoord verticalSpacingIn)
  {
    verticalSpacing = verticalSpacingIn;
  }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58CornerSpacingSpacing2DConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("58 Corner spacing spacing 2D h {} v {}",
                   horizontalSpacing,
                   verticalSpacing);
  }

 protected:
  frCoord horizontalSpacing = -1, verticalSpacing = -1;
};

class frLef58RectOnlyConstraint : public frConstraint
{
 public:
  // constructor
  frLef58RectOnlyConstraint(bool exceptNonCorePinsIn = false)
      : exceptNonCorePins(exceptNonCorePinsIn)
  {
  }
  // getter
  bool isExceptNonCorePinsIn() { return exceptNonCorePins; }
  // setter
  void setExceptNonCorePins(bool exceptNonCorePinsIn)
  {
    exceptNonCorePins = exceptNonCorePinsIn;
  }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58RectOnlyConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("RECTONLY exceptNonCorePins {}", exceptNonCorePins);
  }

 protected:
  bool exceptNonCorePins;
};

class frLef58RightWayOnGridOnlyConstraint : public frConstraint
{
 public:
  // constructor
  frLef58RightWayOnGridOnlyConstraint(bool checkMaskIn = false)
      : checkMask(checkMaskIn)
  {
  }
  // getter
  bool isCheckMask() { return checkMask; }
  // setter
  void setCheckMask(bool checkMaskIn) { checkMask = checkMaskIn; }

  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58RightWayOnGridOnlyConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("RIGHTWAYONGRIDONLY checkMask {}", checkMask);
  }

 protected:
  bool checkMask;
};

class frMetalWidthViaConstraint : public frConstraint
{
 public:
  frMetalWidthViaConstraint(odb::dbMetalWidthViaMap* rule) : dbRule(rule) {}
  odb::dbMetalWidthViaMap* getDbRule() const { return dbRule; }
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcMetalWidthViaConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("METALWIDTHVIAMAP");
  }

 private:
  odb::dbMetalWidthViaMap* dbRule;
};

class frLef58AreaConstraint : public frConstraint
{
 public:
  // constructor
  frLef58AreaConstraint(odb::dbTechLayerAreaRule* dbRule) : db_rule_(dbRule) {}
  // getter
  odb::dbTechLayerAreaRule* getODBRule() const { return db_rule_; }

  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58AreaConstraint;
  }

  void report(utl::Logger* logger) const override
  {
    auto trim_layer = db_rule_->getTrimLayer();
    std::string trim_layer_name
        = trim_layer != nullptr ? db_rule_->getTrimLayer()->getName() : "";
    logger->report(
        "LEF58 AREA rule: area {}, exceptMinWidth {}, exceptEdgeLength {}, "
        "exceptEdgeLengths ({} {}), exceptMinSize ({} {}), exceptStep ({} {}), "
        "mask {}, rectWidth {}, exceptRectangle {}, overlap {}, trimLayer {}",
        db_rule_->getArea(),
        db_rule_->getExceptMinWidth(),
        db_rule_->getExceptEdgeLength(),
        db_rule_->getExceptEdgeLengths().first,
        db_rule_->getExceptEdgeLengths().second,
        db_rule_->getExceptMinSize().first,
        db_rule_->getExceptMinSize().second,
        db_rule_->getExceptStep().first,
        db_rule_->getExceptStep().second,
        db_rule_->getMask(),
        db_rule_->getRectWidth(),
        db_rule_->isExceptRectangle(),
        db_rule_->getOverlap(),
        trim_layer_name);
  }

 private:
  odb::dbTechLayerAreaRule* db_rule_;
};

// LEF58_KEEPOUTZONE rule
class frLef58KeepOutZoneConstraint : public frConstraint
{
 public:
  frLef58KeepOutZoneConstraint(odb::dbTechLayerKeepOutZoneRule* rule)
      : db_rule_(rule)
  {
  }
  // getter
  odb::dbTechLayerKeepOutZoneRule* getODBRule() const { return db_rule_; }
  // others
  frConstraintTypeEnum typeId() const override
  {
    return frConstraintTypeEnum::frcLef58KeepOutZoneConstraint;
  }
  void report(utl::Logger* logger) const override
  {
    logger->report("LEF58_KEEPOUTZONE");
  }

 private:
  odb::dbTechLayerKeepOutZoneRule* db_rule_;
};
class frNonDefaultRule
{
  friend class FlexRP;
  friend class frTechObject;

 private:
  std::string name_;
  // each vector position is a metal layer
  std::vector<frCoord> widths_;
  std::vector<frCoord> spacings_;
  std::vector<frCoord> wireExtensions_;
  std::vector<drEolSpacingConstraint> drEolCons_;
  std::vector<int> minCuts_;  // min cuts per cut layer

  // vias for each layer
  std::vector<std::vector<frViaDef*>> vias_;
  std::vector<std::vector<frViaRuleGenerate*>> viasRules_;

  bool hardSpacing_ = false;

  // See comments in frTechObject's equivalent fields for the meaning
  std::vector<std::array<ForbiddenRanges, 8>> via2ViaForbiddenLen;
  std::vector<std::array<ForbiddenRanges, 4>> viaForbiddenTurnLen;

 public:
  frViaDef* getPrefVia(int z)
  {
    if (z >= (int) vias_.size() || vias_[z].empty()) {
      return nullptr;
    }
    return vias_[z][0];
  }

  void setWidth(frCoord w, int z)
  {
    if (z >= (int) widths_.size()) {
      widths_.resize(z + 1, 0);
    }
    widths_[z] = w;
  }

  void setSpacing(frCoord s, int z)
  {
    if (z >= (int) spacings_.size()) {
      spacings_.resize(z + 1, 0);
    }
    spacings_[z] = s;
  }

  void setMinCuts(int ncuts, int z)
  {
    if (z >= (int) minCuts_.size()) {
      minCuts_.resize(z + 1, 1);
    }
    minCuts_[z] = ncuts;
  }

  void setWireExtension(frCoord we, int z)
  {
    if (z >= (int) wireExtensions_.size()) {
      wireExtensions_.resize(z + 1, 0);
    }
    wireExtensions_[z] = we;
  }

  void setDrEolConstraint(drEolSpacingConstraint con, int z)
  {
    if (z >= (int) drEolCons_.size()) {
      drEolCons_.resize(z + 1, 0);
    }
    drEolCons_[z] = con;
  }

  void addVia(frViaDef* via, int z)
  {
    if (z >= (int) vias_.size()) {
      vias_.resize(z + 1, std::vector<frViaDef*>());
    }
    vias_[z].push_back(via);
  }

  void addViaRule(frViaRuleGenerate* via, int z)
  {
    if (z >= (int) viasRules_.size()) {
      viasRules_.resize(z + 1, std::vector<frViaRuleGenerate*>());
    }
    viasRules_[z].push_back(via);
  }

  void setHardSpacing(bool isHard) { hardSpacing_ = isHard; }

  void setName(const char* n) { name_ = std::string(n); }

  std::string getName() const { return name_; }

  frCoord getWidth(int z) const
  {
    if (z >= (int) widths_.size()) {
      return 0;
    }
    return widths_[z];
  }

  frCoord getSpacing(int z) const
  {
    if (z >= (int) spacings_.size()) {
      return 0;
    }
    return spacings_[z];
  }

  frCoord getWireExtension(int z) const
  {
    if (z >= (int) wireExtensions_.size()) {
      return 0;
    }
    return wireExtensions_[z];
  }

  drEolSpacingConstraint getDrEolSpacingConstraint(int z) const
  {
    if (z >= (int) drEolCons_.size()) {
      return drEolSpacingConstraint();
    }
    return drEolCons_[z];
  }

  int getMinCuts(int z) const
  {
    if (z >= (int) minCuts_.size()) {
      return 1;
    }
    return minCuts_[z];
  }

  const std::vector<frViaDef*>& getVias(int z) const { return vias_[z]; }

  const std::vector<frViaRuleGenerate*>& getViaRules(int z) const
  {
    return viasRules_[z];
  }

  bool isHardSpacing() const { return hardSpacing_; }
};
}  // namespace fr
