///////////////////////////////////////////////////////////////////////////////
// BSD 3-Clause License
//
// Copyright (c) 2019, The Regents of the University of California
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// * Redistributions of source code must retain the above copyright notice, this
//   list of conditions and the following disclaimer.
//
// * Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
//
// * Neither the name of the copyright holder nor the names of its
//   contributors may be used to endorse or promote products derived from
//   this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#pragma once

#include <QAbstractItemDelegate>
#include <QAbstractTableModel>
#include <QAction>
#include <QDockWidget>
#include <QItemDelegate>
#include <QMainWindow>
#include <QMenu>
#include <QModelIndex>
#include <QPoint>
#include <QShortcut>
#include <QSortFilterProxyModel>
#include <QStringList>
#include <QStyledItemDelegate>
#include <QToolBar>
#include <QVariant>
#include <unordered_map>

#include "gui/gui.h"
#include "ui_selectedWidget.h"

namespace gui {
class SelectionModel : public QAbstractTableModel
{
  Q_OBJECT
 public:
  SelectionModel(const SelectionSet& objs);

  int rowCount(const QModelIndex& parent = QModelIndex()) const Q_DECL_OVERRIDE;
  int columnCount(const QModelIndex& parent
                  = QModelIndex()) const Q_DECL_OVERRIDE;

  QVariant data(const QModelIndex& index,
                int role = Qt::DisplayRole) const Q_DECL_OVERRIDE;
  QVariant headerData(int section,
                      Qt::Orientation orientation,
                      int role = Qt::DisplayRole) const Q_DECL_OVERRIDE;

  const Selected* getItemAt(int idx) const { return table_data_[idx]; }

  void populateModel();

 private:
  const SelectionSet& objs_;
  std::vector<const Selected*> table_data_;
};

class HighlightModel : public QAbstractTableModel
{
  Q_OBJECT
 public:
  HighlightModel(const HighlightSet& objs);

  int rowCount(const QModelIndex& parent = QModelIndex()) const Q_DECL_OVERRIDE;
  int columnCount(const QModelIndex& parent
                  = QModelIndex()) const Q_DECL_OVERRIDE;

  QVariant data(const QModelIndex& index,
                int role = Qt::DisplayRole) const Q_DECL_OVERRIDE;
  QVariant headerData(int section,
                      Qt::Orientation orientation,
                      int role = Qt::DisplayRole) const Q_DECL_OVERRIDE;

  const Selected* getItemAt(int idx) const { return table_data_[idx].second; }
  void populateModel();

  int highlightGroup(const QModelIndex& index) const;
  bool setData(const QModelIndex& index,
               const QVariant& value,
               int role) override;

 private:
  const HighlightSet& objs_;
  std::vector<std::pair<int, const Selected*>> table_data_;
};

class SelectHighlightWindow : public QDockWidget
{
  Q_OBJECT

 public:
  explicit SelectHighlightWindow(const SelectionSet& selSet,
                                 const HighlightSet& hltSet,
                                 QWidget* parent = nullptr);
  ~SelectHighlightWindow();

 signals:
  void clearAllSelections();
  void clearAllHighlights();

  void selected(const Selected& selection);
  void clearSelectedItems(const QList<const Selected*>& items);
  void clearHighlightedItems(const QList<const Selected*>& items);
  void zoomInToItems(const QList<const Selected*>& items);
  void highlightSelectedItemsSig(const QList<const Selected*>& items);

 public slots:
  void updateSelectionModel();
  void updateHighlightModel();
  void updateModels();
  void showSelectCustomMenu(QPoint pos);
  void showHighlightCustomMenu(QPoint pos);

  void changeHighlight();

  void deselectItems();
  void highlightSelectedItems();
  void zoomInSelectedItems();
  void dehighlightItems();
  void zoomInHighlightedItems();

 private:
  Ui::SelectHighlightWidget ui_;
  SelectionModel selection_model_;
  QSortFilterProxyModel* sel_filter_proxy_;

  HighlightModel highlight_model_;
  QSortFilterProxyModel* hlt_filter_proxy_;

  QMenu* select_context_menu_;
  QMenu* highlight_context_menu_;
};

}  // namespace gui
